package org.xbib.multiserv.plugin.jdbc;

import org.xbib.jdbc.pool.api.XbibDataSource;
import org.xbib.jdbc.pool.api.configuration.supplier.XbibConnectionFactoryConfigurationSupplier;
import org.xbib.jdbc.pool.api.configuration.supplier.XbibConnectionPoolConfigurationSupplier;
import org.xbib.jdbc.pool.api.configuration.supplier.XbibDataSourceConfigurationSupplier;
import org.xbib.jdbc.pool.api.security.NamePrincipal;
import org.xbib.jdbc.pool.api.security.SimplePassword;
import org.xbib.multiserv.api.Plugin;

import java.io.IOException;
import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.text.Normalizer;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.FormatStyle;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.logging.Level;
import java.util.logging.Logger;

public abstract class JDBCPlugin implements Plugin {

    private static final Logger logger = Logger.getLogger(JDBCPlugin.class.getName());

    private static final AtomicBoolean started = new AtomicBoolean(false);

    private final Properties properties;

    private static XbibDataSource poolDataSource;

    public JDBCPlugin() {
        Properties systemProperties = System.getProperties();
        String url = systemProperties.getProperty("multiserv.jdbc.url");
        String username = systemProperties.getProperty("multiserv.jdbc.username");
        String password = systemProperties.getProperty("multiserv.jdbc.password");
        this.properties = new Properties();
        if (url != null) {
            properties.put("url", url);
        }
        if (username != null) {
            properties.put("username", username);
        }
        if (password != null) {
            properties.put("password", password);
        }
    }

    public synchronized void onStart() throws IOException {
        if (started.compareAndSet(false, true)) {
            try {
                NetworkUtils.waitForNetworkInterfaces();
                logger.log(Level.INFO, "network is up");
                poolDataSource = createDataSource(properties);
                logger.log(Level.INFO, "pool is up");
            } catch (Exception e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }
    }

    private XbibDataSource createDataSource(Properties properties) throws SQLException {
        XbibDataSourceConfigurationSupplier supplier = new XbibDataSourceConfigurationSupplier();
        XbibConnectionPoolConfigurationSupplier poolConfig = supplier.connectionPoolConfiguration();
        poolConfig.maxSize(4);
        XbibConnectionFactoryConfigurationSupplier connectionFactoryConfig = poolConfig.connectionFactoryConfiguration();
        connectionFactoryConfig.jdbcUrl((String) properties.get("url"));
        if (properties.containsKey("username")) {
            connectionFactoryConfig.credential(new NamePrincipal((String) properties.get("username")));
        }
        if (properties.containsKey("password")) {
            connectionFactoryConfig.credential(new SimplePassword((String) properties.get("password")));
        }
        return  XbibDataSource.from(supplier.get());
    }

    @Override
    public synchronized void onShutdown() {
        if (started.compareAndSet(true, false)) {
            if (poolDataSource != null) {
                try {
                    poolDataSource.close();
                    poolDataSource = null;
                    logger.log(Level.INFO, "pool is down");
                } catch (Exception e) {
                    logger.log(Level.SEVERE, e.getMessage(), e);
                }
            }
        }
    }

    protected Map<String, Object> getRow(String statement, Map<String, Object> params) throws SQLException {
        List<Map<String, Object>> rows = getRows(statement, params);
        return rows.isEmpty() ? new LinkedHashMap<>() : rows.getFirst();
    }

    protected List<Map<String, Object>> getRows(String statement, Map<String, Object> params) throws SQLException {
        List<Map<String, Object>> rows = new ArrayList<>();
        if (params == null || params.isEmpty()) {
            queryAndApply(statement, Map.of(), r -> {
                try {
                    while (r.next()) {
                        rows.add(getNextRow(r));
                    }
                } catch (SQLException t) {
                    logger.log(Level.SEVERE, t.getMessage(), t);
                }
            });
        } else {
            queryAndApply(statement, params, r -> {
                try {
                    while (r.next()) {
                        rows.add(getNextRow(r));
                    }
                } catch (SQLException t) {
                    logger.log(Level.SEVERE, t.getMessage(), t);
                }
            });
        }
        return rows;
    }

    protected void executeStatement(String query, Map<String, Object> params) throws SQLException {
        try (Connection connection = poolDataSource.getConnection();
             NamedParameterStatement statement = new NamedParameterStatement(connection, query, false)) {
            for (Map.Entry<String, Object> entry : params.entrySet()) {
                statement.setObject(entry.getKey(), entry.getValue());
            }
            statement.execute();
        }
    }

    protected void insertStatement(String query,
                                   Map<String, Object> params) throws SQLException {
        try (Connection connection = poolDataSource.getConnection();
             NamedParameterStatement statement = new NamedParameterStatement(connection, query, false)) {
            for (Map.Entry<String, Object> entry : params.entrySet()) {
                statement.setObject(entry.getKey(), entry.getValue());
            }
            statement.executeUpdate();
        }
    }

    protected void insertStatement(String query,
                                   List<Map<String, Object>> paramList) throws SQLException {
        try (Connection connection = poolDataSource.getConnection();
             NamedParameterStatement statement = new NamedParameterStatement(connection, query, false)) {
            for (Map<String, Object> params : paramList) {
                for (Map.Entry<String, Object> entry : params.entrySet()) {
                    statement.setObject(entry.getKey(), entry.getValue());
                }
                statement.addBatch();
            }
            statement.executeBatch();
        }
    }

    protected void upsertStatement(String deleteQuery,
                                   List<Map<String, Object>> deleteParamList,
                                   String insertQuery,
                                   List<Map<String, Object>> insertParamList) throws SQLException {
        try (Connection connection = poolDataSource.getConnection();
             NamedParameterStatement deleteStatement = new NamedParameterStatement(connection, deleteQuery, false, false);
             NamedParameterStatement insertStatement = new NamedParameterStatement(connection, insertQuery, false, false)) {
            for (Map<String, Object> params : deleteParamList) {
                for (Map.Entry<String, Object> entry : params.entrySet()) {
                    deleteStatement.setObject(entry.getKey(), entry.getValue());
                }
                deleteStatement.addBatch();
            }
            deleteStatement.executeBatch();
            for (Map<String, Object> params : insertParamList) {
                for (Map.Entry<String, Object> entry : params.entrySet()) {
                    insertStatement.setObject(entry.getKey(), entry.getValue());
                }
                insertStatement.addBatch();
            }
            insertStatement.executeBatch();
        }
    }

    protected void queryAndApply(String query,
                               Map<String, Object> params,
                               Consumer<ResultSet> consumer) throws SQLException {
        try (Connection connection = poolDataSource.getConnection();
             NamedParameterStatement statement = new NamedParameterStatement(connection, query, true)) {
            for (Map.Entry<String, Object> entry : params.entrySet()) {
                statement.setObject(entry.getKey(), entry.getValue());
            }
            consumer.accept(statement.executeQuery());
        }
    }

    private static Map<String, Object> getNextRow(ResultSet r) throws SQLException {
        Map<String, Object> result = new LinkedHashMap<>();
        // JDBC columns start at 1
        for (int i = 1; i <= r.getMetaData().getColumnCount(); i++) {
            String k = r.getMetaData().getColumnLabel(i).toLowerCase(Locale.ROOT);
            Object v = r.getObject(i);
            if (v instanceof Timestamp) {
                v = ((Timestamp) v).toLocalDateTime();
            }
            if (k.endsWith("time")) {
                v = formatDateTime((String) v);
            } else if (v instanceof String) {
                // normalize decomposed UTF and trim eventual CHAR columns
                v = Normalizer.normalize((String) v, Normalizer.Form.NFC).trim();
            }
            result.put(k, v);
        }
        return result;
    }

    private static String formatDateTime(String instant) {
        return ZonedDateTime.parse(instant).toLocalDateTime().format(shortFormatter);
    }

    private static final DateTimeFormatter shortFormatter =
            DateTimeFormatter.ofLocalizedDateTime(FormatStyle.SHORT, FormatStyle.SHORT)
                    .withLocale(Locale.getDefault())
                    .withZone(ZoneId.systemDefault());
}
