package org.xbib.multiserv.plugin.jdbc;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public class NamedParameterStatement implements AutoCloseable {

    private final Connection connection;

    private final boolean isAutoCommit;

    private final PreparedStatement statement;

    private final Map<String, List<Integer>> paramMap;

    public NamedParameterStatement(Connection connection,
                                   String query) throws SQLException {
        this(connection, query, true, true);
    }

    public NamedParameterStatement(Connection connection,
                                   String query,
                                   boolean isFetch) throws SQLException {
        this(connection, query, isFetch, true);
    }

    public NamedParameterStatement(Connection connection,
                                   String query,
                                   boolean isFetch,
                                   boolean isAutoCommit) throws SQLException {
        this.connection = connection;
        this.isAutoCommit = isAutoCommit;
        connection.setAutoCommit(isAutoCommit);
        paramMap = new LinkedHashMap<>();
        String parsedQuery = parse(query);
        if (isFetch) {
            statement = connection.prepareStatement(parsedQuery,
                    ResultSet.TYPE_FORWARD_ONLY,
                    ResultSet.CONCUR_READ_ONLY);
            statement.setFetchDirection(ResultSet.FETCH_FORWARD);
            statement.setFetchSize(100);
        } else {
            statement = connection.prepareStatement(parsedQuery);
        }
    }

    @Override
    public void close() throws SQLException {
        if (!isAutoCommit) {
            connection.commit();
        }
        statement.close();
    }

    public void setObject(String name, Object value) throws SQLException {
        List<Integer> indexes = paramMap.get(name);
        if (indexes != null) {
            for (Integer index : indexes) {
                statement.setObject(index, value);
            }
        }
    }

    public void execute() throws SQLException {
        statement.execute();
    }

    public void addBatch() throws SQLException {
        statement.addBatch();
    }

    public void executeBatch() throws SQLException {
        statement.executeBatch();
    }

    public void executeUpdate() throws SQLException {
        statement.executeUpdate();
    }

    public ResultSet executeQuery() throws SQLException {
        return statement.executeQuery();
    }

    private String parse(String query) {
        int length = query.length();
        StringBuilder parsedQuery = new StringBuilder(length);
        boolean inSingleQuote = false;
        boolean inDoubleQuote = false;
        int index = 1;
        for (int i = 0; i < length; i++) {
            char c = query.charAt(i);
            if (inSingleQuote) {
                if (c == '\'') {
                    inSingleQuote = false;
                }
            } else if (inDoubleQuote) {
                if (c == '"') {
                    inDoubleQuote = false;
                }
            } else {
                if (c == '\'') {
                    inSingleQuote = true;
                } else if (c == '"') {
                    inDoubleQuote = true;
                } else if (c == ':' && i + 1 < length && Character.isJavaIdentifierStart(query.charAt(i + 1))) {
                    int j = i + 2;
                    while (j < length && Character.isJavaIdentifierPart(query.charAt(j))) {
                        j++;
                    }
                    String name = query.substring(i + 1, j);
                    c = '?';
                    i += name.length();
                    List<Integer> indexList = paramMap.computeIfAbsent(name, k -> new ArrayList<>());
                    indexList.add(index);
                    index++;
                }
            }
            parsedQuery.append(c);
        }
        return parsedQuery.toString();
    }
}
